from functools import lru_cache
from django.utils.module_loading import import_string
import random
import string


__all__ = (
    'cached_import_string',
    'make_unique_label',
)


@lru_cache
def cached_import_string(path: str):
    return import_string(path)


def make_unique_label(length: int, alphabet: str = string.ascii_lowercase) -> str:
    return ''.join(random.choices(alphabet, k=length))



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy

from .const import APP_LABEL


__all__ = ('MassAffectConfig',)


class MassAffectConfig(AppConfig):
    name = 'pxd_massaffect'
    label = APP_LABEL
    verbose_name = pgettext_lazy('pxd_massaffect', 'MassAffect')
    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self) -> None:
        super().ready()

        self.module.autodiscover()



from django.utils.translation import pgettext_lazy
from django.db import models


class ExcludableConfig(models.Model):
    class Meta:
        abstract = True

    exclude = models.BooleanField(
        verbose_name=pgettext_lazy('pxd_massaffect', 'Exclude'),
        default=False, null=False,
    )


class ExcludableAdapterMixin:
    def exclude_Q(self, config: ExcludableConfig, q: models.Q) -> models.Q:
        return ~q if config.exclude else q



from dataclasses import dataclass
from px_settings.contrib.django import settings as s


__all__ = 'Settings', 'settings',


@s('pxd_massaffect')
@dataclass
class Settings:
    pass


settings = Settings()



from typing import Union, Optional, Type, TypeVar
from django.db import models
from django.utils.translation import pgettext_lazy
from polymorphic.models import PolymorphicModel


__all__ = (
    'GenericConfig',
    'make_formula_field',
    'make_config_base',
)

T = TypeVar('T')


class GenericConfig(PolymorphicModel):
    variable_prefix = 'affect'

    class Meta:
        abstract = True
        verbose_name = pgettext_lazy('pxd_massaffect', 'Config')
        verbose_name_plural = pgettext_lazy('pxd_massaffect', 'Configs')
        ordering = ('-ordering_field', '-pk',)

    parent = models.ForeignKey(
        'content_type.ContentType', on_delete=models.CASCADE
    )
    ordering_field = models.BigIntegerField(
        verbose_name=pgettext_lazy('pxd_massaffect', 'Ordering value'),
        db_index=True, default=0,
    )

    @property
    def variable_name(self):
        return self.variable_prefix + str(self.pk)

    def __str__(self):
        return self.variable_name


def make_formula_field(**kwargs) -> models.TextField:
    return models.TextField(
        verbose_name=pgettext_lazy('pxd_massaffect', 'Formula'),
        blank=True, null=False, **kwargs,
    )


def make_config_base(
    name: str,
    parent: Union[str, Type[models.Model]],
    /,
    base: T = GenericConfig,
    related_name: str = 'affect_configs',
    app_label: Optional[str] = None,
    **kwargs
) -> T:
    return type(name.title() + base.__name__, (base,), {
        '__module__': __name__, '__doc__': base.__doc__,
        'Meta': type('Meta', (base.Meta,), {
            'abstract': True,
            'app_label': app_label,
            'verbose_name': base._meta.verbose_name,
            'verbose_name_plural': base._meta.verbose_name_plural,
        }),
        'parent': models.ForeignKey(
            parent,
            verbose_name=pgettext_lazy('pxd_massaffect', 'Parent'),
            related_name=related_name,
            on_delete=models.CASCADE, null=False, blank=False,
            **kwargs,
        )
    })



__version__ = '0.1.0'

from django.utils.module_loading import autodiscover_modules


VERSION = tuple(__version__.split('.'))

default_app_config = 'pxd_massaffect.apps.SalesConfig'


def autodiscover():
    autodiscover_modules('sales')



from typing import Type, TYPE_CHECKING
import nested_admin

from django.db import models

if TYPE_CHECKING:
    from .registry import AffectRegistry


__all__ = 'AffectConfigSetInline', 'AffectConfigInline', 'make_inline'


class AffectConfigSetInline(
    nested_admin.SortableHiddenMixin,
    nested_admin.NestedStackedPolymorphicInline,
):
    registry: 'AffectRegistry'

    model = None
    sortable_field_name = 'ordering_field'

    def get_registry_child_inlines(self):
        return [x.admin_inline for x in self.registry.values()]

    def get_child_inline_instances(self):
        inlines = list(self.child_inlines) + self.get_registry_child_inlines()

        instances = []
        for ChildInlineType in inlines:
            instances.append(ChildInlineType(parent_inline=self))

        return instances


class AffectConfigInline(
    nested_admin.SortableHiddenMixin,
    nested_admin.NestedStackedPolymorphicInline.Child
):
    model = None
    sortable_field_name = 'ordering_field'


def make_inline(
    model: Type[models.Model],
    /,
    base = AffectConfigInline,
) -> AffectConfigInline:
    return type(model.__name__ + 'Inline', (base,), {'model': model})



from typing import List, Optional, Sequence, Tuple, Union
from collections import OrderedDict

from .functional import cached_import_string


__all__ = 'Adapter', 'Descriptor', 'DescriptorRegistry',


class Adapter:
    pass


class Descriptor:
    key: str
    verbose_name: str
    adapter: Adapter

    def __init__(
        self,
        key: str,
        adapter: Adapter,
        verbose_name: Optional[str] = None,
    ):
        self.key = key
        self.adapter = adapter
        self.verbose_name = (
            verbose_name if verbose_name is not None else key.title()
        )


class DescriptorRegistry(OrderedDict):
    def register(self, descriptor: Descriptor):
        assert not self.registered(descriptor.key), f'{descriptor.key} already registered.'

        self[descriptor.key] = descriptor

    def registered(self, name: str) -> bool:
        return name in self

    def choices(self) -> List[Tuple[str, str]]:
        return [
            (key, descriptor.verbose_name)
            for key, descriptor in self.items()
        ]

    def multiregister(self, descriptors: Sequence[Union[str, Descriptor]]):
        for descriptor in descriptors:
            self.register(
                cached_import_string(descriptor)
                if isinstance(descriptor, str) else
                descriptor
            )



from functools import cached_property, reduce
from collections import OrderedDict
import operator
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, TypeVar, Union
from django.db import models

from px_formula import Formula

from .models import GenericConfig
from .registry import AffectRegistry, FieldGeneratorMixin


__all__ = (
    'and_resolver', 'or_resolver', 'make_formula_resolver',
    'Affector', 'ParentsAffector',
)

Q = TypeVar('Q', bound=models.QuerySet)
T = TypeVar('T')
M = TypeVar('M', bound=models.Model)
Variables = Dict[str, Union[models.F, Any]]


def and_resolver(variables: Variables):
    return reduce(operator.and_, variables.values()) if variables else None


def or_resolver(variables: Variables):
    return reduce(operator.or_, variables.values()) if variables else None


def make_formula_resolver(
    formula: Formula,
    expression: str,
    fallback: Callable[[Variables], Any] = and_resolver,
    merge: Callable = operator.and_,
):
    if not expression:
        return fallback

    parsed = formula.parse(expression)
    searchable = set(parsed)

    def resolver(variables: Variables):
        diff = variables.keys() - searchable
        result = formula.evaluate(parsed, variables)

        if len(diff) == 0:
            return result

        return merge(result, fallback({k: variables[k] for k in diff}))

    return resolver


class Affector(FieldGeneratorMixin):
    registry: AffectRegistry

    def __init__(self, registry: AffectRegistry):
        self.registry = registry

    def affect_queryset(
        self,
        configs: Sequence[GenericConfig],
        queryset: Q,
        resolver: Callable[[Variables], Any] = and_resolver,
        kwargs = {},
    ) -> Tuple[Q, str]:
        result: Variables = {}
        registry = self.registry
        field_name = self.make_unique_name()
        model = queryset.model

        for config in configs:
            config_model = config.__class__
            descriptor = registry.get_for_model(config_model)

            if (
                descriptor is not None
                and
                descriptor.adapter.can_affect(model)
            ):
                queryset, field = descriptor.adapter.affect_queryset(
                    config, queryset, kwargs=kwargs
                )
                result[config.variable_name] = models.ExpressionWrapper(
                    models.F(field), output_field=models.BooleanField(),
                )

        return (
            queryset.annotate(**{
                field_name: resolver(result) if result else models.Value(False),
            }),
            field_name,
        )

    def affect_items(
        self,
        configs: Sequence[GenericConfig],
        items: Sequence[T],
        resolver: Callable[[Variables], Any] = and_resolver,
        kwargs = {},
    ) -> List[Tuple[T, Any]]:
        registry = self.registry
        result: Dict[int, Variables] = {x: {} for x in range(len(items))}
        items = list(items)

        for config in configs:
            config_model = config.__class__
            descriptor = registry.get_for_model(config_model)

            if descriptor is None:
                continue

            affectable = OrderedDict(
                (i, item) for i, item in enumerate(items)
                if descriptor.adapter.can_affect(item.__class__)
            )

            if len(affectable) == 0:
                continue

            indexes = list(affectable.keys())
            temp = descriptor.adapter.affect_items(
                config, affectable.values(), kwargs=kwargs,
            )
            variable_name = config.variable_name

            for i, (item, value) in enumerate(temp):
                index = indexes[i]
                result[index][variable_name] = value
                items[index] = item

        return [(item, resolver(result[i])) for i, item in enumerate(items)]


class ParentsAffector:
    parents: Sequence[M]
    config_queryset: models.QuerySet = None
    related_name: str = 'affect_configs'
    affector: Affector = None

    def __init__(
        self,
        parents: Sequence[M],
        config_queryset: models.QuerySet = None,
        related_name: Optional[str] = None,
        affector: Optional[Affector] = None,
        get_parent_resolver: Optional[Callable[[M], Callable[[Variables], Any]]] = None,
    ):
        assert len(parents) > 0, 'There must be at leas one parent passed'

        self.affector = self.affector if affector is None else affector
        self.config_queryset = self.config_queryset if config_queryset is None else config_queryset
        self.related_name = self.related_name if related_name is None else related_name

        for value, name in (
            (self.affector, 'Affector'),
            (self.related_name, 'Related name'),
        ):
            assert value is not None, (
                f'{name} must be either declared as class property or passed on '
                'initialization.'
            )

        self.parents = parents

        if self.config_queryset is None:
            field = getattr(parents[0].__class__, self.related_name)
            self.config_queryset = field.rel.related_model.objects.all()

        if get_parent_resolver is not None:
            self.get_parent_resolver = get_parent_resolver

    def get_parent_resolver(self, parent: M) -> Callable[[Variables], Any]:
        return and_resolver

    @cached_property
    def parent_configs(self) -> Dict[M, List[GenericConfig]]:
        queryset = self.config_queryset

        if callable(queryset):
            queryset = queryset()

        configs = queryset.filter(parent__in=self.parents)
        result = OrderedDict((parent, []) for parent in self.parents)
        parents_map = {x.pk: x for x in self.parents}

        for config in configs:
            parent = parents_map[config.parent_id]
            result[parent].append(config)

        return result

    def affect_queryset(
        self,
        queryset: Q,
        get_parent_resolver: Optional[Callable[[M], Callable[[Variables], Any]]] = None,
        kwargs = {},
    ) -> Tuple[Q, Dict[M, str]]:
        parent_configs = self.parent_configs
        affector = self.affector
        results = {}
        get_parent_resolver = (
            self.get_parent_resolver
            if get_parent_resolver is None else
            get_parent_resolver
        )

        for parent, configs in parent_configs.items():
            queryset, field_name = affector.affect_queryset(
                configs, queryset, resolver=get_parent_resolver(parent),
                kwargs=kwargs,
            )
            results[parent] = field_name

        return queryset, results

    def affect_items(
        self,
        items: Sequence[T],
        get_parent_resolver: Optional[Callable[[M], Callable[[Variables], Any]]] = None,
        kwargs = {},
    ) -> List[Tuple[T, Dict[M, Any]]]:
        parent_configs = self.parent_configs
        affector = self.affector
        results = [[x, {}] for x in items]
        get_parent_resolver = (
            self.get_parent_resolver
            if get_parent_resolver is None else
            get_parent_resolver
        )

        for parent, configs in parent_configs.items():
            temp = affector.affect_items(
                configs, items, resolver=get_parent_resolver(parent),
                kwargs=kwargs,
            )
            for i, (item, value) in enumerate(temp):
                results[i][0] = item
                results[i][1][parent] = value

        return results



from typing import Any, List, Optional, Sequence, Tuple, Type, TypeVar
from django.db import models

from .admin import AffectConfigInline, make_inline
from .utils.registry import Adapter, Descriptor, DescriptorRegistry
from .utils import make_unique_label, get_model_content_type_key


Q = TypeVar('Q')
T = TypeVar('T')


class FieldGeneratorMixin:
    field_prefix: str = 'affected_'

    def make_unique_name(self, prefix: Optional[str] = None):
        prefix = prefix if prefix is not None else self.field_prefix

        return prefix + make_unique_label(8)


class AffectAdapter(FieldGeneratorMixin, Adapter):
    affected_models: Sequence[Type[models.Model]]

    def __init__(self, affected_models: Sequence[Type[models.Model]]):
        self.affected_models = affected_models

    def can_affect(self, model: Type[models.Model]):
        return model in self.affected_models

    def affect_Q(self, config: models.Model, **kwargs) -> models.Q:
        raise NotImplementedError()

    def affect_queryset(
        self,
        config: models.Model,
        queryset: Q,
        kwargs = {},
    ) -> Tuple[Q, str]:
        field_name = self.make_unique_name()
        q = self.affect_Q(config, **kwargs)
        sub_q = models.Exists(queryset.filter(q).filter(pk=models.OuterRef('pk')))

        return queryset.annotate(**{field_name: q}), field_name

    def affect_item(self, config: models.Model, item: T, kwargs = {}) -> Any:
        raise NotImplementedError()

    def affect_items(
        self,
        config: models.Model,
        items: Sequence[T],
        kwargs = {},
    ) -> List[Tuple[T, Any]]:
        return [
            (item, self.affect_item(config, item, kwargs=kwargs))
            for item in items
        ]


class AffectDescriptor(Descriptor):
    model: Type[models.Model]
    adapter: AffectAdapter
    admin_inline: Type[AffectConfigInline]

    _make_inline = staticmethod(make_inline)

    def __init__(
        self,
        model: Type[models.Model],
        adapter: AffectAdapter,
        verbose_name: Optional[str] = None,
        admin_inline: Optional[Type[AffectConfigInline]] = None,
    ):
        self.model = model
        self.admin_inline = (
            admin_inline if admin_inline is not None else self._make_inline(model)
        )
        opts = model._meta

        super().__init__(
            get_model_content_type_key(model),
            adapter=adapter,
            verbose_name=(
                verbose_name if verbose_name is not None else
                opts.verbose_name
            )
        )


class AffectRegistry(DescriptorRegistry):
    def get_for_model(self, model: models.Model, dflt=None) -> Optional[AffectDescriptor]:
        return self.get(get_model_content_type_key(model), dflt)



from typing import Callable

from django.db import models


__all__ = 'get_model_content_type_key',


def get_model_content_type_key(model: models.Model) -> str:
    opts = model._meta
    return '.'.join((opts.app_label, opts.model_name))
